from concurrent.futures import ThreadPoolExecutor
from functools import reduce
import itertools
import logging
from datetime import datetime
from queue import Empty, Queue
import time
import timeit
from typing import List, Callable, Iterable, Iterator, TypeVar

T = TypeVar("T")
R = TypeVar("R")


class Pipe(Iterator[T]):
    def __init__(self, iterator: Iterator[T]) -> None:
        self.iterator: Iterator[T] = iter(
            iterator
        )  # call iter in case it is an iterable that is passed

    def __next__(self) -> T:
        return next(self.iterator)

    def __iter__(self) -> T:
        return self

    def chain(self, *others: Iterable["Pipe[T]"]) -> "Pipe[T]":
        return Pipe(itertools.chain(self, *others))

    def merge(self, *others: Iterable["Pipe[T]"]) -> "Pipe[T]":
        return _ConcurrentlyMergingPipe([self, *others])

    def map(self, func: Callable[[T], R], num_threads: int = 0) -> "Pipe[R]":
        if num_threads <= 1:
            return Pipe(map(func, self))
        else:

            def iterable():
                executor = ThreadPoolExecutor(max_workers=num_threads)
                yield from executor.map(func, self)
                executor.shutdown()

            return Pipe(iter(iterable()))

    def filter(self, predicate: Callable[[T], bool]) -> "Pipe[R]":
        return Pipe(filter(predicate, self))

    def batch(
        self, max_size: int = 100, max_window_seconds: float = float("inf")
    ) -> "Pipe[List[T]]":
        return _BatchingPipe(self, max_size, max_window_seconds)

    def rate_limit(self, max_yields_per_second: int) -> "Pipe[T]":
        return Pipe(_RateLimitingIterator(self, max_yields_per_second))

    def head(self, n) -> "Pipe[T]":
        return Pipe(itertools.islice(self, n))

    def catch(self) -> "Pipe[T]":
        return _CatchingPipe(self)

    def log(self) -> "Pipe[T]":
        return _LoggingPipe(self)

    def reduce(self, f: Callable[[R, T], R], initial: R) -> R:
        return reduce(f, self, initial)

    def exhaust_into_limited_list(self, limit: int = float("inf")) -> List[T]:
        return [elem for i, elem in enumerate(self.iterator) if i < limit]

    def timeit(self) -> float:
        def iterate():
            for _ in self:
                pass

        return timeit.timeit(iterate, number=1)


class CatchedPipeError:
    def __init__(self, e: Exception) -> None:
        self.error_type = type(e)
        self.message = str(e)

    def __str__(self) -> str:
        return f"{self.error_type}: {self.message}"


class _CatchingPipe(Pipe[T]):
    def __next__(self) -> T:
        try:
            return super().__next__()
        except Exception as e:
            if isinstance(e, StopIteration):
                raise
            else:
                return CatchedPipeError(e)


class _LoggingPipe(Pipe[T]):
    def __init__(self, iterator: Iterator[T]) -> None:
        super().__init__(iterator)
        logging.getLogger().setLevel(logging.INFO)
        self.yields_count = 0
        self.errors_count = 0
        self.last_log_at_yields_count = 0
        self.start_time = time.time()

    def _log(self) -> None:
        logging.info(
            "%s yields in %s, leading to the catch of %s errors.",
            self.yields_count,
            str(
                datetime.fromtimestamp(time.time())
                - datetime.fromtimestamp(self.start_time)
            ),
            self.errors_count,
        )

    def __next__(self) -> T:
        try:
            elem = super().__next__()
        except StopIteration:
            self._log()
            raise

        self.yields_count += 1
        if isinstance(elem, CatchedPipeError):
            self.errors_count += 1

        if self.yields_count + self.errors_count >= 2 * self.last_log_at_yields_count:
            self._log()
            self.last_log_at_yields_count = self.yields_count + self.errors_count

        return elem


class _RateLimitingIterator(Pipe[T]):
    def __init__(self, iterator: Iterator[T], max_yields_per_second: int) -> None:
        super().__init__(iterator)
        self.max_yields_per_second = max_yields_per_second
        self.start = None
        self.num_yielded_elements = 0

    def __next__(self) -> T:
        if not self.start:
            self.start = time.time()
        while True:
            while (
                self.num_yielded_elements
                > (time.time() - self.start) * self.max_yields_per_second
            ):
                time.sleep(1 / self.max_yields_per_second)
            self.num_yielded_elements += 1
            return super().__next__()


class _BatchingPipe(Pipe[List[T]]):
    def __init__(
        self, iterator: Iterator[T], max_size: int, max_window_seconds: float
    ) -> None:
        super().__init__(iterator)
        self.max_size = max_size
        self.max_window_seconds = max_window_seconds

    def __next__(self) -> List[T]:
        start_time = time.time()
        batch = [next(self.iterator)]
        try:
            while (
                len(batch) < self.max_size
                and (time.time() - start_time) < self.max_window_seconds
            ):
                batch.append(next(self.iterator))
            return batch
        except StopIteration:
            if batch:
                return batch
            else:
                raise


class _ConcurrentlyMergingPipe(Pipe[T]):
    def __init__(self, iterators: List[Iterator[T]]) -> None:
        super().__init__(
            iter(_ConcurrentlyMergingPipe._concurrently_merging_iterable(iterators))
        )

    @staticmethod
    def _pull_in_queue(iterator: Iterator[T], queue: Queue) -> None:
        for elem in iterator:
            queue.put(elem)

    @staticmethod
    def _concurrently_merging_iterable(iterators: List[Iterator[T]]) -> Iterator[T]:
        queue = Queue()
        with ThreadPoolExecutor(max_workers=len(iterators)) as executor:
            futures = [
                executor.submit(
                    _ConcurrentlyMergingPipe._pull_in_queue, iterator, queue
                )
                for iterator in iterators
            ]
            while not queue.empty() or not all((future.done() for future in futures)):
                try:
                    yield queue.get(block=False)
                except Empty:
                    time.sleep(0.05)
